"""NAME
    Multicall

DESCRIPTION
    A multicall for use with pure Web3 lbrary.
    It uses MakerDao Multicall smart contract by default,
    but also can use any custom multicall smart contract 
    that implements "aggregate" function

"""

from web3 import Web3
from web3.eth import Contract

from .makerdao_multicall import (
    MAKERDAO_MULTICALL_ABI, 
    MAKERDAO_MULTICALL_ADDRESS
)


class Multicall:
    """
    NAME
        Multicall
    
    DESCRIPTION
       The main multicall class.
    
    ATTRIBUTES
        w3 : Web3 class instanse

        chain: str
            The name of one of defined chains 
            where MakerDao Multicall smart contract is deployed.
            Can be one of the followings:
            - 'mainnet'
            - 'kovan'
            - 'rinkeby'
            - 'goerli'
            - 'ropsten'
            - 'xdai'
            - 'polygon'
            - 'mumbai'
            - 'bsc-mainnet'
            - 'bsc-testnet'

        custom_address: str
            An address of custom multicall smart contract. 
            If specified, MakerDao Multicall smart contract will be omited.

        custom_abi: str
            An ABI of custom multicall smart contract.
            If omited, MakerDao Multicall smart contract ABI will be used.

    """

    def __init__(
        self,
        w3: Web3,
        chain='mainnet',
        custom_address=None,
        custom_abi=None
    ):
        if custom_address:
            address = Web3.toChecksumAddress(custom_address)
        else:
            try:
                address = Web3.toChecksumAddress(MAKERDAO_MULTICALL_ADDRESS[chain])  
            except(Exception):
                print('Chain name key not in default dictionary')
                return  

        abi=MAKERDAO_MULTICALL_ABI        

        if custom_abi:
            abi = custom_abi

        self.multicall = w3.eth.contract(
            address=address,
            abi=abi
        )


    def call(
        self,
        calls: list
    ) -> list:
        """
        Executes multicall for specified list of smart contracts functions.

        Parameters:
            calls: list(tuple)
                list of tuples (target contract address, encoded function name with parameters)
                Can be easy prepared via using Multicall.create_cal' method.

        Returns:
            tuple(block number, list(results)) for default MakerDao multicall. May vary for custom multicalls.
        """
        
        return self.multicall.functions.aggregate(calls).call()


    def create_call(
        self,
        contract: Contract,
        fn_name: str,
        args: list
    ) -> tuple:
        """
        Prepares a tuple for passing to Multicall.call list.

        Parameters:
            contract: Web3.eth.Contract
                A Web3.eth.Contract instance of a contract 
                that is to be called via multicall.

            fn_name: str
                The name of a contract function to be called.

            args: list
                a list of arguments to be passed to a called contract function.

        Returns:
            tuple(target contract address, encoded function name with parameters)

        """

        return (contract.address, contract.encodeABI(fn_name=fn_name, args=args))
        


from setuptools import setup, find_packages
from pathlib import Path

VERSION = '0.1.4' 
DESCRIPTION = 'Simple Web3 multicall'

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='simple_multicall',
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Igor Lapshin',
    author_email= 'igorlapshin@list.ru',
    packages=find_packages(),
    license='MIT',
    keywords=['multicall', 'web3'],
    requires=['Web3'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3"
    ]
)


MAKERDAO_MULTICALL_ABI = '[{"constant":true,"inputs":[],"name":"getCurrentBlockTimestamp","outputs":[{"name":"timestamp","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"components":[{"name":"target","type":"address"},{"name":"callData","type":"bytes"}],"name":"calls","type":"tuple[]"}],"name":"aggregate","outputs":[{"name":"blockNumber","type":"uint256"},{"name":"returnData","type":"bytes[]"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"getLastBlockHash","outputs":[{"name":"blockHash","type":"bytes32"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"addr","type":"address"}],"name":"getEthBalance","outputs":[{"name":"balance","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"getCurrentBlockDifficulty","outputs":[{"name":"difficulty","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"getCurrentBlockGasLimit","outputs":[{"name":"gaslimit","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"getCurrentBlockCoinbase","outputs":[{"name":"coinbase","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"blockNumber","type":"uint256"}],"name":"getBlockHash","outputs":[{"name":"blockHash","type":"bytes32"}],"payable":false,"stateMutability":"view","type":"function"}]'

MAKERDAO_MULTICALL_ADDRESS = {
    'mainnet': '0xeefba1e63905ef1d7acba5a8513c70307c1ce441',
    'kovan': '0x2cc8688c5f75e365aaeeb4ea8d6a480405a48d2a',
    'rinkeby': '0x42ad527de7d4e9d9d011ac45b31d8551f8fe9821',
    'goerli': '0x77dca2c955b15e9de4dbbcf1246b4b85b651e50e',
    'ropsten': '0x53c43764255c17bd724f74c4ef150724ac50a3ed',
    'xdai': '0xb5b692a88bdfc81ca69dcb1d924f59f0413a602a',
    'polygon': '0x11ce4B23bD875D7F5C6a31084f55fDe1e9A87507',
    'mumbai': '0x08411ADd0b5AA8ee47563b146743C13b3556c9Cc',
    'bsc-mainnet': '0x41263cba59eb80dc200f3e2544eda4ed6a90e76c',
    'bsc-testnet': '0xae11C5B5f29A6a25e955F0CB8ddCc416f522AF5C'
}